/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */
#pragma warning(disable: 4312)

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#define WINDOWS_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <cufft.h>
#include <cutil.h>
#include <cuda_d3d9_interop.h>
#include <d3dx9.h>

#include "fluidsD3D_kernels.cu"

// CUDA example code that implements the frequency space version of 
// Jos Stam's paper 'Stable Fluids' in 2D. This application uses the 
// CUDA FFT library (CUFFT) to perform velocity diffusion and to 
// force non-divergence in the velocity field at each time step. It uses 
// CUDA-OpenGL interoperability to update the particle field directly
// instead of doing a copy to system memory before drawing. Texture is
// used for automatic bilinear interpolation at the velocity advection step. 

#ifdef __DEVICE_EMULATION__
#define DIM    64        // Square size of solver domain
#else
#define DIM    512       // Square size of solver domain
#endif
#define DS    (DIM*DIM)  // Total domain size
#define CPADW (DIM/2+1)     // Padded width for real->complex in-place FFT
#define RPADW (2*(DIM/2+1)) // Padded width for real->complex in-place FFT
#define PDS   (DIM*CPADW)   // Padded total domain size

#define DT     0.09f     // Delta T for interative solver
#define VIS    0.0025f   // Viscosity constant
#define FORCE (5.8f*DIM) // Force scale factor 
#define FR     4         // Force update radius

#define TILEX 64 // Tile width
#define TILEY 64 // Tile height
#define TIDSX 64 // Tids in X
#define TIDSY 4  // Tids in Y

HWND                    hWnd;              // Window handle 
LPDIRECT3D9             pD3D       = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       pd3dDevice = NULL; // Rendering device
LPDIRECT3DVERTEXBUFFER9 pVB        = NULL; // Buffer to hold particles
LPDIRECT3DTEXTURE9      pTexture   = NULL; // Texture to render points

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE)
void updateVB(void);
void initParticles(cData *p, int dx, int dy);

// CUFFT plan handle
static cufftHandle planr2c;
static cufftHandle planc2r;
static cData *vxfield = NULL;
static cData *vyfield = NULL;

cData *hvfield = NULL;
cData *dvfield = NULL;
static int wWidth = max(512,DIM);
static int wHeight = max(512,DIM);

static int clicked = 0;
static int fpsCount = 0;
static int fpsLimit = 1;
unsigned int timer;

// Particle data
static Vertex *mparticles = NULL;  
static cData *particles = NULL;
static int lastx = 0, lasty = 0;

// Texture pitch
//unsigned int tPitch = 0;
size_t tPitch = 0;

void addForces(cData *v, int dx, int dy, int spx, int spy, float fx, float fy, int r) { 

    uint3 tids = make_uint3(2*r+1, 2*r+1, 1);
    
    addForces_k<<<1, tids>>>(v, dx, dy, spx, spy, fx, fy, r, tPitch);
    CUT_CHECK_ERROR("addForces_k failed.");
}

void advectVelocity(cData *v, float *vx, float *vy,
                    int dx, int pdx, int dy, float dt) { 
    
    uint3 grid = make_uint3((dx/TILEX)+(!(dx%TILEX)?0:1), 
                            (dy/TILEY)+(!(dy%TILEY)?0:1), 1);

    uint3 tids = make_uint3(TIDSX, TIDSY, 1);

    updateTexture(v, DIM*sizeof(cData), DIM, tPitch);
    advectVelocity_k<<<grid, tids>>>(v, vx, vy, dx, pdx, dy, dt, TILEY/TIDSY);

    CUT_CHECK_ERROR("advectVelocity_k failed.");
}

void diffuseProject(cData *vx, cData *vy, int dx, int dy, float dt,
                    float visc) { 
    // Forward FFT
    cufftExecR2C(planr2c, (cufftReal*)vx, (cufftComplex*)vx); 
    cufftExecR2C(planr2c, (cufftReal*)vy, (cufftComplex*)vy);

    uint3 grid = make_uint3((dx/TILEX)+(!(dx%TILEX)?0:1), 
                            (dy/TILEY)+(!(dy%TILEY)?0:1), 1);

    uint3 tids = make_uint3(TIDSX, TIDSY, 1);
    
    diffuseProject_k<<<grid, tids>>>(vx, vy, dx, dy, dt, visc, TILEY/TIDSY);
    CUT_CHECK_ERROR("diffuseProject_k failed.");

    // Inverse FFT
    cufftExecC2R(planc2r, (cufftComplex*)vx, (cufftReal*)vx); 
    cufftExecC2R(planc2r, (cufftComplex*)vy, (cufftReal*)vy);
}

void updateVelocity(cData *v, float *vx, float *vy, 
                    int dx, int pdx, int dy) { 

    uint3 grid = make_uint3((dx/TILEX)+(!(dx%TILEX)?0:1), 
                            (dy/TILEY)+(!(dy%TILEY)?0:1), 1);

    uint3 tids = make_uint3(TIDSX, TIDSY, 1);

    updateVelocity_k<<<grid, tids>>>(v, vx, vy, dx, pdx, dy, TILEY/TIDSY, tPitch);
    CUT_CHECK_ERROR("updateVelocity_k failed.");
}

void advectParticles(Vertex *p, cData *v, int dx, int dy, float dt) {
    
    uint3 grid = make_uint3((dx/TILEX)+(!(dx%TILEX)?0:1), 
                            (dy/TILEY)+(!(dy%TILEY)?0:1), 1);

    uint3 tids = make_uint3(TIDSX, TIDSY, 1);

    cudaD3D9MapVertexBuffer((void**)&p, pVB);
    CUT_CHECK_ERROR("cudaD3D9MapVertexBuffer failed");
   
    advectParticles_k<<<grid, tids>>>(p, v, dx, dy, dt, TILEY/TIDSY, tPitch);
    CUT_CHECK_ERROR("advectParticles_k failed.");
    
    cudaD3D9UnmapVertexBuffer(pVB); 
    CUT_CHECK_ERROR("cudaD3D9UnmapVertexBuffer failed");
}

VOID Cleanup() { 
    unbindTexture();
    deleteTexture();
    
    // Free all host and device resources
    free(hvfield); free(particles); 
    cudaFree(dvfield); 
    cudaFree(vxfield); cudaFree(vyfield);
    cufftDestroy(planr2c);
    cufftDestroy(planc2r);
    
	cudaD3D9End();
    CUT_CHECK_ERROR("cudaD3D9End failed"); 
    
    if (pVB != NULL) {        
        pVB->Release();
        pVB = NULL;
    }
    if (pTexture != NULL) {        
        pTexture->Release();
        pTexture = NULL;
    }
    if (pd3dDevice != NULL) {
        pd3dDevice->Release();
        pd3dDevice = NULL;
    }
    if (pD3D != NULL) {       
        pD3D->Release();  
        pD3D = NULL;
    }  
    CUT_SAFE_CALL(cutDeleteTimer(timer));  
}

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    switch (msg) {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage(0);
            return 0;
        case WM_KEYDOWN:
            switch(wParam) {
                case 27:
                Cleanup();
                PostQuitMessage(0);
                break;
                case 0x52:
                    memset(hvfield, 0, sizeof(cData) * DS);
                    cudaMemcpy(dvfield, hvfield, sizeof(cData) * DS, 
                               cudaMemcpyHostToDevice);

                    initParticles(particles, DIM, DIM);
                    updateVB();
                    break;
                default:
                break;
            }
            break;
        case WM_MOUSEMOVE: 
            if (wParam == MK_LBUTTON) clicked = 1;
            else clicked = 0;
            int x = LOWORD(lParam), y = HIWORD(lParam);
             
            // Convert motion coordinates to domain
            float fx = (x / (float)wWidth);        
            float fy = (y / (float)wHeight);
            int nx = (int)(fx * DIM);        
            int ny = (int)(fy * DIM);   
           
            if (clicked && nx < DIM-FR && nx > FR-1 && ny < DIM-FR && ny > FR-1) {
                int ddx = LOWORD(lParam) - lastx;
                int ddy = HIWORD(lParam) - lasty;
                
                fx = ddx / (float)wWidth;
                fy = ddy / (float)wHeight;
                int spy = ny-FR;
                int spx = nx-FR;
                addForces(dvfield, DIM, DIM, spx, spy, FORCE * DT * fx, FORCE * DT * fy, FR);
                lastx = x; lasty = y;
            } 
            break;
    }
    return DefWindowProc( hWnd, msg, wParam, lParam );
}

HRESULT InitVB() {
    // Create the vertex buffer.
    if (FAILED(pd3dDevice->CreateVertexBuffer(DS * sizeof(Vertex),
                                              0, D3DFVF_CUSTOMVERTEX,
                                              D3DPOOL_DEFAULT, &pVB, NULL)))
        return E_FAIL;
    
    updateVB();
    
    return S_OK;
}

HRESULT InitPointTexture() {
    // Create the texture.
    int width = 64;
    int height = width;
    if (FAILED(D3DXCreateTexture(pd3dDevice,
                                 width, height, D3DX_DEFAULT,
                                 D3DUSAGE_AUTOGENMIPMAP, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED,
                                 &pTexture)))
        return E_FAIL;

    // Fill in top level
    D3DLOCKED_RECT rect;
    if (FAILED(pTexture->LockRect(0, &rect, 0, 0)))
        return E_FAIL;
    typedef unsigned int TexelType;
    TexelType* texel = (TexelType*)rect.pBits;
    for (int y = - height / 2; y < height / 2; ++y) {
        float yf = y + 0.5f;
        TexelType* t = texel;
        for (int x = - width / 2; x < width / 2; ++x) {
            float xf = x + 0.5f;
            float radius = (float)width / 32;
            float dist = sqrtf(xf * xf + yf * yf) / radius;
            float n = 0.1f;
            float value;
            if (dist < 1)
                value = 1 - 0.5f * powf(dist, n);
            else if (dist < 2)
                value = 0.5f * powf(2 - dist, n);
            else
                value = 0;
            value *= 75;
            unsigned char* c = (unsigned char*)t;
            c[0] = c[1] = c[2] = c[3] = (unsigned char)value;
            ++t;
        }
        texel += rect.Pitch / sizeof(TexelType);
    }
    if (FAILED(pTexture->UnlockRect(0)))
        return E_FAIL;

    // Set sampler state
    if (FAILED(pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR)))
        return E_FAIL;
    if (FAILED(pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR)))
        return E_FAIL;
    
    return S_OK;
}

void updateVB(void) {
    Vertex *data = NULL;
    pVB->Lock(0,DS * sizeof(Vertex),(void**)&data,0);
    for (int i = 0; i < DS; i++) {
        data[i].x = particles[i].x;
        data[i].y = particles[i].y;
        data[i].z = 0.f;
        data[i].c = 0xff00ff00;
    }
    pVB->Unlock();
}

HRESULT InitD3D(HWND hWnd) {
    // Create the D3D object.
    if(NULL == (pD3D = Direct3DCreate9(D3D_SDK_VERSION)))
        return E_FAIL;
    
    // Get primary display identifier
    D3DADAPTER_IDENTIFIER9 adapterId;
    pD3D->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &adapterId);
    
    // Direct3D and CUDA can only interoperate on the same device
    int device;
    if (cudaSuccess == cudaD3D9GetDevice(&device, adapterId.DeviceName))
        CUDA_SAFE_CALL(cudaSetDevice(device));
    else {
        printf("The primary display is not CUDA-capable.\n");
        return E_FAIL;
    }

    // Set up the structure used to create the D3DDevice
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    
    // Create the D3DDevice
    if ( FAILED( pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                     D3DCREATE_HARDWARE_VERTEXPROCESSING,
                                     &d3dpp, &pd3dDevice ) ) )
        return E_FAIL;    
       
    // Set projection matrix
    D3DXMATRIXA16 matProj;
    D3DXMatrixOrthoOffCenterLH( &matProj, 0, 1, 1, 0, 0, 1);
    pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );
   
    // Turn off D3D lighting, since we are providing our own vertex colors
    if( FAILED( pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE ) ) )
        return E_FAIL;
        
    return S_OK;
}

void display(void) {  
   CUT_SAFE_CALL(cutStartTimer(timer));  
    
   advectVelocity(dvfield, (float*)vxfield, (float*)vyfield, DIM, RPADW, DIM, DT);
   diffuseProject(vxfield, vyfield, CPADW, DIM, DT, VIS);
   updateVelocity(dvfield, (float*)vxfield, (float*)vyfield, DIM, RPADW, DIM);
   advectParticles(mparticles, dvfield, DIM, DIM, DT);
   
   pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
   pd3dDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
   pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
   pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
   pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
   pd3dDevice->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
   float size = 16;
   pd3dDevice->SetRenderState(D3DRS_POINTSIZE, *((DWORD*)&size));
   pd3dDevice->SetTexture(0, pTexture);

   if(SUCCEEDED(pd3dDevice->BeginScene())) {
   
       // Draw particles
       pd3dDevice->SetStreamSource(0, pVB, 0, sizeof(Vertex));
       pd3dDevice->SetFVF(D3DFVF_CUSTOMVERTEX);
       pd3dDevice->DrawPrimitive(D3DPT_POINTLIST, 0, DS);
    
       pd3dDevice->EndScene();
   }

   // Finish timing before swap buffers to avoid refresh sync
   CUT_SAFE_CALL(cutStopTimer(timer));  
   // Present the backbuffer contents to the display
   pd3dDevice->Present(NULL, NULL, NULL, NULL);

   fpsCount++;
   if (fpsCount == fpsLimit) {
       char fps[256];
       float ifps = 1.f / (cutGetAverageTimerValue(timer) / 1000.f);
       sprintf(fps, "Cuda/D3D Stable Fluids (%d x %d): %3.1f fps", DIM, DIM, ifps);  
       SetWindowText(hWnd, fps); 
       fpsCount = 0; 
       fpsLimit = (int)max(ifps, 1.f);
       CUT_SAFE_CALL(cutResetTimer(timer));  
    }
}

void initParticles(cData *p, int dx, int dy) {
    int i, j;
    for (i = 0; i < dy; i++) {
        for (j = 0; j < dx; j++) {
            p[i*dx+j].x = ((j+0.5f)/dx) + 
                          (rand() / (float)RAND_MAX - 0.5f) / dx;
            p[i*dx+j].y = ((i+0.5f)/dy) + 
                          (rand() / (float)RAND_MAX - 0.5f) / dy;
                          
        }
    }
}

int main(int argc, char** argv) {
    CUT_SAFE_CALL(cutCreateTimer(&timer));
    CUT_SAFE_CALL(cutResetTimer(timer));  
    
    hvfield = (cData*)malloc(sizeof(cData) * DS);
    memset(hvfield, 0, sizeof(cData) * DS);
  
    // Allocate and initialize device data
    cudaMallocPitch((void**)&dvfield, &tPitch, sizeof(cData)*DIM, DIM);
    
    cudaMemcpy(dvfield, hvfield, sizeof(cData) * DS, 
               cudaMemcpyHostToDevice); 
    // Temporary complex velocity field data     
    cudaMalloc((void**)&vxfield, sizeof(cData) * PDS);
    cudaMalloc((void**)&vyfield, sizeof(cData) * PDS);
    
    setupTexture(DIM, DIM);
    bindTexture();
    
    // Create particle array
    particles = (cData*)malloc(sizeof(cData) * DS);
    memset(particles, 0, sizeof(cData) * DS);   
    
    initParticles(particles, DIM, DIM); 

    // Create CUFFT transform plan configuration
    cufftPlan2d(&planr2c, DIM, DIM, CUFFT_R2C);
    cufftPlan2d(&planc2r, DIM, DIM, CUFFT_C2R);
    
    HINSTANCE hInst = GetModuleHandle(NULL);
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "fluidsD3D", NULL };
    RegisterClassEx(&wc);

    // Create the application's window
    hWnd = CreateWindow("fluidsD3D", "Cuda/D3D Stable Fluids",
                        WS_OVERLAPPEDWINDOW, 100, 100, wWidth, wHeight,
                        NULL, NULL, wc.hInstance, NULL );

    if (SUCCEEDED(InitD3D(hWnd)) && SUCCEEDED(InitVB()) && SUCCEEDED(InitPointTexture())) {
        cudaD3D9Begin(pd3dDevice);
        CUT_CHECK_ERROR("cudaD3D9Begin failed");
        cudaD3D9RegisterVertexBuffer(pVB);
        CUT_CHECK_ERROR("cuD3D9RegisterVertexBuffer failed");

        ShowWindow(hWnd, SW_SHOWDEFAULT);
        UpdateWindow(hWnd);

        // Rendering loop
        MSG msg;
        ZeroMemory(&msg, sizeof(msg));
        while(msg.message!=WM_QUIT) {

            if(PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            } else { display();    }
        }
    } 
    UnregisterClass("fluidsD3D", wc.hInstance );
}
